#define PROCMON_EXTERN
#define PROCMON_EXTERN_OFF
#include "procmon.h"
#if defined( _PROCMON_DEBUG_EXTRA )
static int num_times_global;
#endif
/* ========================================================================

   Name
   ====
   procmon_info_init

   Purpose
   =======
   Initialize data before sampling starts

   Arguments
   =========
   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) process environment variables and set variables

   ======================================================================== */
int procmon_info_init(
                      void
                      )
{
  int
    ierr=0; /* error return value */
  long long
    part_sec, /* part of a second (some units like microseconds) */
    secs; /* number of seconds */
  char
    *getenv_char_p, /* pointer to getenv command char array */
    *this_dir = ".", /* current working directory */
    *filename_start = "procmon_", /* start of filename */
    *filename_separator = "_", /* separates fields in filename */
    filename_pid[32], /* string version of pid */
    *filename_end = "XXXXXX"; /* end of filename - Xs for use in mkstemp */
  
  /* --------------- */
  /* set some things */
  /* --------------- */
  procmon_s.time_init_coarse = time(NULL);
  if( clock_gettime( CLOCK_REALTIME, &(procmon_s.time_init_fine) ) != 0 )
    abort();
  procmon_s.pid = getpid();
  /* ----------- */
  /* set actions */
  /* ----------- */
  memset(&(procmon_s.action_old),0x00,sizeof(struct sigaction));
  memset(&(procmon_s.action),0x00,sizeof(struct sigaction));
  procmon_s.action.sa_flags = SA_RESTART|SA_SIGINFO;
  procmon_s.action.sa_sigaction = procmon_handler;
  memset(&(procmon_s.action_ignore),0x00,sizeof(struct sigaction));
  procmon_s.action_ignore.sa_handler = SIG_IGN;
  /* ---------------- */
  /* set itimer value */
  /* ---------------- */
  procmon_s.millisecs = 0;
  getenv_char_p = getenv( "PROCMON_TIME" );
  if( getenv_char_p != NULL )
    {
      procmon_s.millisecs = atoi( getenv_char_p );
      if( procmon_s.millisecs < 0 )
        {
          procmon_s.millisecs = 0;
        }
    }
  part_sec = (procmon_s.millisecs%1000)*1000;
  secs = procmon_s.millisecs/1000;
  procmon_s.value.it_interval.tv_sec = secs;
  procmon_s.value.it_interval.tv_usec = part_sec;
  procmon_s.value.it_value.tv_sec = secs;
  procmon_s.value.it_value.tv_usec = part_sec;
  /* --------------------- */
  /* set print_sample_data */
  /* --------------------- */
  procmon_s.output_interval = 0;
  getenv_char_p = getenv( "PROCMON_INTERVAL" );
  if( getenv_char_p != NULL )
    {
      procmon_s.output_interval = atoi( getenv_char_p );
      if( procmon_s.output_interval < 0 )
        {
          procmon_s.output_interval = 0;
        }
    }
  /* ------------------------- */
  /* set print_file_sample_min */
  /* ------------------------- */
  getenv_char_p = getenv( "PROCMON_MIN" );
  procmon_s.output_sample_min = 0;
  if( getenv_char_p != NULL )
    {
      procmon_s.output_sample_min = atoi( getenv_char_p );
      if( procmon_s.output_sample_min < 0 )
        {
          procmon_s.output_sample_min = 0;
        }
    }
  /* ------------------- */
  /* set output filename */
  /* ------------------- */
  getenv_char_p = getenv( "PROCMON_DIR" );
  if( getenv_char_p == NULL || strlen( getenv_char_p ) <= 0 )
    {
      getenv_char_p = this_dir;
    }
  if(gethostname(procmon_s.hostname, PROCMON_MAXHOSTNAMELEN-1) != 0)
    abort();
  sprintf( filename_pid, "%d", procmon_s.pid );
  procmon_s.filename =
    (char*)malloc( strlen( getenv_char_p ) +
                   1 +
                   strlen( filename_start ) +
                   strlen( procmon_s.hostname ) +
                   strlen( filename_separator ) +
                   strlen( filename_pid ) +
                   strlen( filename_separator ) +
                   strlen( filename_end ) + 1 );
  sprintf( procmon_s.filename, "%s%s%s%s%s%s%s%s",
           getenv_char_p, "/", filename_start, procmon_s.hostname,
           filename_separator, filename_pid, filename_separator,
           filename_end );
  /* ---------------------------- */
  /* open file and print a header */
  /* ---------------------------- */
  /* ---------------------------------------- */
  /* no need to mkstemp...just overwrite file */
  /* ---------------------------------------- */
  /*
    procmon_s.filename_fd = mkstemp( procmon_s.filename );
    if( procmon_s.filename_fd == -1 )
    abort();
    if( close( procmon_s.filename_fd ) != 0 )
    abort();
  */
  procmon_s.filename_fp = fopen( procmon_s.filename, "a" );
#if defined( _PROCMON_DEBUG_EXTRA )
  fprintf( stderr, "extra %s ns %lld ms %lld open gv %d\n",
           procmon_s.filename, procmon_s.sample_info.num_samples,
           procmon_s.output_sample_min, num_times_global );
  fflush( stderr );
#endif
  if( procmon_s.filename_fp != NULL )
    {
      fprintf( procmon_s.filename_fp,
               "# %40s: %s\n"
               "# %40s: %llu\n"
               "# %40s: %llu\n"
               "# %40s: %llu\n"
               "# %40s: %s",
               "Filename", procmon_s.filename,
               "PROCMON_TIME", procmon_s.millisecs,
               "PROCMON_MIN", procmon_s.output_sample_min,
               "PROCMON_INTERVAL", procmon_s.output_interval,
               "Creation Time", asctime(localtime(&(procmon_s.time_init_coarse)))
               );
      fflush( procmon_s.filename_fp );
    }
  procmon_s.initialized = !0;
  return( ierr );
}
/* ========================================================================

   Name
   ====
   procmon_info_free

   Purpose
   =======
   free stuff gotten in init call

   Arguments
   =========
   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) process environment variables and set variables

   ======================================================================== */
int procmon_info_free(
                      void
                      )
{
  int
    ierr=0; /* error return value */
#if defined( _PROCMON_DEBUG_EXTRA )
  static int num_times_local=0;
#endif
  assert( procmon_s.initialized != 0 );
  procmon_s.initialized = 0;
#if defined( _PROCMON_DEBUG_EXTRA )
  num_times_local++;
#endif
  if( procmon_s.filename_fp != NULL )
    {
#if defined( _PROCMON_DEBUG_EXTRA )
      fprintf( stderr, "extra %s ns %lld sm %lld close gv %d lv %d \n",
               procmon_s.filename, procmon_s.sample_info.num_samples,
               procmon_s.output_sample_min, num_times_global,
               num_times_local );
      fflush( stderr );
#endif
      ierr = fclose( procmon_s.filename_fp );
      assert( !ierr );
      procmon_s.filename_fp = NULL;
      /* --------------------------------- */
      /* delete file if not enough samples */
      /* --------------------------------- */
      if( procmon_s.sample_info.num_samples <
          procmon_s.output_sample_min )
        {
#if defined( _PROCMON_DEBUG_EXTRA )
          fprintf( stderr, "extra %s ns %lld sm %lld delete gv %d lv %d\n",
                   procmon_s.filename,
                   procmon_s.sample_info.num_samples,
                   procmon_s.output_sample_min,
                   num_times_global, num_times_local );
          fflush( stderr );
#endif
          ierr = remove( procmon_s.filename );
          assert( !ierr );
        }
    }
  free( procmon_s.filename );
  procmon_s.filename = NULL;
  return( ierr );
}
/* ========================================================================

   Name
   ====
   procmon_handler

   Purpose
   =======
   Get/print process info info

   Arguments
   =========
   sig           Intent: in
                 C type: int
                 The signal sent to this process.
                 If called independently, the value should be set to 0
                 Currently not used, but could be in the future.

   info          Intent: in
                 C type: siginfo_t *
                 One of the required args for sa_handler

   context       Intent: in
                 C type: void *
                 One of the required args for sa_handler

   Return Values
   =============
   none

   Program Flow
   ============
   1) get info
   2) print info

   ======================================================================== */
void procmon_handler(
                     int sig,
                     siginfo_t *info,
                     void *context
                     )
{
  struct rusage
    rusage_info; /* getrusage( RUSAGE_[SELF|CHILDERN], &rusage_info ) */
#if defined( _PROCMON_ALPHA ) | defined( _PROCMON_IRIX )
  /* see /usr/include/sys/procfs.h */
  prpsinfo_t
    prpsinfo_info; /* ioctl(fd, PIOCPSINFO, &prpsinfo_info) */
  prusage_t
    prusage_info; /* ioctl(fd, PIOCUSAGE,  &prusage_info) */
  char
    pfile[256]; /* proc file */
  int
    fd; /* proc file handle */
#endif
  int
    i; /* loop var */
  struct
    mallinfo mi; /* mallinfo() */
#if defined( _PROCMON_DEBUG_EXTRA )
  num_times_global++;
#endif
  /* ------------------------------------------------------------- */
  /* set to ignore SIGPROF (do not want to get signals backing up) */
  /* ------------------------------------------------------------- */
  if( sig != 0 )
    {
      if( sigaction(SIGPROF, &(procmon_s.action_ignore), NULL ) != 0 )
          abort();
    }
  /* ----------- */
  /* assign data */
  /* ----------- */
  /* ------------- */
  /* clock_gettime */
  /* ------------- */
  if( clock_gettime( CLOCK_REALTIME, &(procmon_s.time_current) ) != 0 )
    abort();
  procmon_s.sample_info.time_data[0] =
    PROCMON_TIMEN_TO_DOUBLE( procmon_s.time_current );
  procmon_s.sample_info.time_data[2] =
    procmon_s.sample_info.time_data[1];
  procmon_s.sample_info.time_data[1] =
    PROCMON_TIMEN_TO_DOUBLE( procmon_s.time_current ) -
    PROCMON_TIMEN_TO_DOUBLE( procmon_s.time_init_fine );
  procmon_s.sample_info.time_data[2] =
    procmon_s.sample_info.time_data[1] -
    procmon_s.sample_info.time_data[2];
  /* ------------------- */
  /* DONE: clock_gettime */
  /* ------------------- */
  /* -------- */
  /* mallinfo */
  /* -------- */
  mi = mallinfo();
  procmon_s.sample_info.page_size = sysconf(_SC_PAGESIZE) ;
  procmon_s.sample_info.int_vector[ 0][1] = mi.arena;
  procmon_s.sample_info.int_vector[ 1][1] = mi.ordblks;
  procmon_s.sample_info.int_vector[ 2][1] = mi.smblks;
  procmon_s.sample_info.int_vector[ 3][1] = mi.hblks;
  procmon_s.sample_info.int_vector[ 4][1] = mi.hblkhd;
  procmon_s.sample_info.int_vector[ 5][1] = mi.usmblks;
  procmon_s.sample_info.int_vector[ 6][1] = mi.fsmblks;
  procmon_s.sample_info.int_vector[ 7][1] = mi.uordblks;
  procmon_s.sample_info.int_vector[ 8][1] = mi.fordblks;
  procmon_s.sample_info.int_vector[ 9][1] = mi.keepcost;
  /* -------------- */
  /* DONE: mallinfo */
  /* -------------- */
  /* --------- */
  /* getrusage */
  /* --------- */
  if( getrusage( RUSAGE_SELF, &rusage_info ) != 0 )
    abort();
  procmon_s.sample_info.int_vector[10][1] = rusage_info.ru_maxrss;
  procmon_s.sample_info.int_vector[11][1] = rusage_info.ru_ixrss;
  procmon_s.sample_info.int_vector[12][1] = rusage_info.ru_idrss;
  procmon_s.sample_info.int_vector[13][1] = rusage_info.ru_isrss;
  procmon_s.sample_info.int_vector[14][1] = rusage_info.ru_minflt;
  procmon_s.sample_info.int_vector[15][1] = rusage_info.ru_majflt;
  procmon_s.sample_info.int_vector[16][1] = rusage_info.ru_nswap;
  procmon_s.sample_info.int_vector[17][1] = rusage_info.ru_inblock;
  procmon_s.sample_info.int_vector[18][1] = rusage_info.ru_oublock;
  procmon_s.sample_info.int_vector[19][1] = rusage_info.ru_msgsnd;
  procmon_s.sample_info.int_vector[20][1] = rusage_info.ru_msgrcv;
  procmon_s.sample_info.int_vector[21][1] = rusage_info.ru_nsignals;
  procmon_s.sample_info.int_vector[22][1] = rusage_info.ru_nvcsw;
  procmon_s.sample_info.int_vector[23][1] = rusage_info.ru_nivcsw;
  procmon_s.sample_info.time_data[3] =
    PROCMON_TIMEU_TO_DOUBLE( rusage_info.ru_utime );
  procmon_s.sample_info.time_data[4] =
    PROCMON_TIMEU_TO_DOUBLE( rusage_info.ru_stime );
  /* --------------- */
  /* DONE: getrusage */
  /* --------------- */
  /* ----- */
  /* ioctl */
  /* ----- */
#if defined( _PROCMON_ALPHA )
  sprintf(pfile, "/proc/%05d", procmon_s.pid);
  fd = open(pfile,O_RDONLY);
  if (ioctl(fd, PIOCPSINFO, &prpsinfo_info)<0)
    abort();
  if (ioctl(fd, PIOCUSAGE,  &prusage_info)<0)
    abort();
  close(fd);
  procmon_s.sample_info.int_vector[24][1] = prpsinfo_info.pr_rssize;
  procmon_s.sample_info.int_vector[25][1] = prpsinfo_info.pr_size;
  procmon_s.sample_info.time_data[5] =
    PROCMON_TIMEN_TO_DOUBLE(prusage_info.pr_wtime);
#elif defined( _PROCMON_IRIX )
  sprintf(pfile, "/proc/%05d", procmon_s.pid);
  fd = open(pfile,O_RDONLY);
  if (ioctl(fd, PIOCPSINFO, &prpsinfo_info)<0)
    abort();
  if (ioctl(fd, PIOCUSAGE,  &prusage_info)<0)
    abort();
  close(fd);
  procmon_s.sample_info.int_vector[24][1] = prpsinfo_info.pr_rssize;
  procmon_s.sample_info.int_vector[25][1] = prpsinfo_info.pr_size;
  procmon_s.sample_info.time_data_ndefined[5] = !0;
#else
  procmon_s.sample_info.int_vector_ndefined[24] = !0;
  procmon_s.sample_info.int_vector_ndefined[25] = !0;
  procmon_s.sample_info.time_data_ndefined[5] = !0;
#endif
  /* ----------- */
  /* DONE: ioctl */
  /* ----------- */
  /* ----------------- */
  /* DONE: assign data */
  /* ----------------- */
  /* ----------- */
  /* set min/max */
  /* ----------- */
  for( i = 0; i < PROCMON_NAME_VALUE_MAX; i++ )
    {
      /* -------- */
      /* finished */
      /* -------- */
      if( strlen( procmon_name_value_int_vector[i].name ) <= 0 )
        {
          break;
        }
      /* ----------- */
      /* not defined */
      /* ----------- */
      if( procmon_s.sample_info.int_vector_ndefined[i] != 0 )
        {
          continue;
        }
      /* ------------- */
      /* initial value */
      /* ------------- */
      if( procmon_s.sample_info.num_samples == 0 )
        {
          procmon_s.sample_info.int_vector[i][0] =
            procmon_s.sample_info.int_vector[i][1];
          procmon_s.sample_info.int_vector[i][2] =
            procmon_s.sample_info.int_vector[i][1];
        }
      /* ------------ */
      /* find min/max */
      /* ------------ */
      else
        {
          procmon_s.sample_info.int_vector[i][0] =
            PROCMON_MIN( procmon_s.sample_info.int_vector[i][1],
                         procmon_s.sample_info.int_vector[i][0] );
          procmon_s.sample_info.int_vector[i][2] =
            PROCMON_MAX( procmon_s.sample_info.int_vector[i][1],
                         procmon_s.sample_info.int_vector[i][2] );
        }
    }
  /* -------------------------------- */
  /* increment number of samples data */
  /* -------------------------------- */
  procmon_s.sample_info.num_samples++;
  procmon_s.sample_info.num_samples_interval++;
  if( procmon_s.sample_info.num_samples_interval ==
      procmon_s.output_interval )
    {
      procmon_s.sample_info.num_samples_interval = 0;
    }
  /* ------------------------ */
  /* print info               */
  /* file exists and          */
  /*  - correct interval (or) */
  /*  - final output          */
  /* ------------------------ */
  if( ( procmon_s.filename_fp != NULL ) &&
      ( ( procmon_s.sample_info.num_samples_interval == 0 ) ||
        ( procmon_s.sample_info.num_samples > 1 && sig == 0 ) ) )
    {
      procmon_s.sample_info.num_samples_printed++;
      /* ------------------- */
      /* print sample header */
      /* ------------------- */
      fprintf( procmon_s.filename_fp, "#\n" );
      fprintf( procmon_s.filename_fp, "# %40s: %llu\n",
               "Sample_number", procmon_s.sample_info.num_samples );
      if( info == NULL )
        {
          fprintf( procmon_s.filename_fp, "# %40s: %s\n",
                   "Sample_type", "auto" );
        }
      else if( info->si_code == SI_QUEUE )
        {
          fprintf( procmon_s.filename_fp, "# %40s: %s\n",
                   "Sample_type", "queue" );
        }
      else if( info->si_code == SI_ASYNCIO )
        {
          fprintf( procmon_s.filename_fp, "# %40s: %s\n",
                   "Sample_type", "asyncio" );
        }
      else if( info->si_code == SI_MESGQ )
        {
          fprintf( procmon_s.filename_fp, "# %40s: %s\n",
                   "Sample_type", "mesgq" );
        }
      else if( info->si_code == SI_USER )
        {
          fprintf( procmon_s.filename_fp, "# %40s: %s\n",
                   "Sample_type", "user" );
        }
      else
        {
          fprintf( procmon_s.filename_fp, "# %40s: %s\n",
                   "Sample_type", "timer" );
        }
      fprintf( procmon_s.filename_fp, "# %40s: %llu\n",
               "Samples_printed",
               procmon_s.sample_info.num_samples_printed );
      for( i = 0; i < PROCMON_NAME_VALUE_MAX; i++ )
        {
          /* -------- */
          /* finished */
          /* -------- */
          if( strlen( procmon_name_value_time_data[i].name ) == 0 )
            {
              break;
            }
          /* ----------- */
          /* not defined */
          /* ----------- */
          if( procmon_s.sample_info.time_data_ndefined[i] != 0 )
            {
              continue;
            }
          fprintf( procmon_s.filename_fp, "# %40s: %17f\n",
                   procmon_name_value_time_data[i].name,                   
                   procmon_s.sample_info.time_data[i] );
        }
      fprintf( procmon_s.filename_fp, "#\n" );
      fprintf( procmon_s.filename_fp, "#+-----------------+" );
      for( i = 0; i < 4; i++ )
        {
          fprintf( procmon_s.filename_fp, "------------+" );
        }
      fprintf( procmon_s.filename_fp, "\n" );
      fprintf( procmon_s.filename_fp, "#|%17s|%12s|%12s|%12s|%12s|\n",
               "Time(sec)", "Min", "Current", "Max", "Type" );
      fprintf( procmon_s.filename_fp, "#+-----------------+" );
      for( i = 0; i < 4; i++ )
        {
          fprintf( procmon_s.filename_fp, "------------+" );
        }
      fprintf( procmon_s.filename_fp, "\n" );
      /* ----------------- */
      /* print sample data */
      /* ----------------- */
      for( i = 0; i < PROCMON_NAME_VALUE_MAX; i++ )
        {
          /* -------- */
          /* finished */
          /* -------- */
          if( strlen( procmon_name_value_int_vector[i].name ) == 0 )
            {
              break;
            }
          /* ----------- */
          /* not defined */
          /* ----------- */
          if( procmon_s.sample_info.int_vector_ndefined[i] != 0 )
            {
              continue;
            }
          fprintf( procmon_s.filename_fp,
                   "  %17f %12.5e %12.5e %12.5e  %s\n",
                   procmon_s.sample_info.time_data[0],
                   (double)procmon_s.sample_info.int_vector[i][0],
                   (double)procmon_s.sample_info.int_vector[i][1],
                   (double)procmon_s.sample_info.int_vector[i][2],
                   procmon_name_value_int_vector[i].name );
        }
      fprintf( procmon_s.filename_fp, "\n\n" );
      fflush( procmon_s.filename_fp );
    }
  /* ---------------- */
  /* DONE: print info */
  /* ---------------- */
  /* --------------------------------------------- */
  /* chain to previous sig catcher if given signal */
  /* --------------------------------------------- */
  if( sig != 0 )
    {
      /* ------------------------------------- */
      /* Do I want to chain???  perhaps not... */
      /* ------------------------------------- */
      /* ------------------------------------------ */
      /* to remove compiler warning if not chaining */
      /* ------------------------------------------ */
      context = context;
      if( procmon_s.action_old.sa_flags & SA_SIGINFO &&
          procmon_s.action_old.sa_sigaction != NULL )
        {
          procmon_s.action_old.sa_sigaction(sig, info, context);
        }
      else if( procmon_s.action_old.sa_handler != NULL )
        {
          procmon_s.action_old.sa_handler(sig);
        }
    }
  /* --------------------------------------------------- */
  /* DONE: chain to previous sig catcher if given signal */
  /* --------------------------------------------------- */
  /* ------------------------ */
  /* no longer ingore SIGPROF */
  /* ------------------------ */
  if( sig != 0 )
    {
      if (sigaction(SIGPROF, &(procmon_s.action), NULL ) != 0 )
        abort();
    }
}
void procmon_ctor(
                  void
                  )
{
  /* --------- */
  /* init info */
  /* --------- */
  if( procmon_info_init() != 0 )
    abort();
  /* ------------------------- */
  /* get sampling at beginning */
  /* ------------------------- */
  procmon_handler(0,NULL,NULL);
  /* ---------------------------------------------------------- */
  /* set catcher first then start itimer in case timer goes off */
  /* ---------------------------------------------------------- */
  if (sigaction(SIGPROF, &(procmon_s.action),
                &(procmon_s.action_old)) != 0)
    abort();
  if (setitimer(ITIMER_PROF, &(procmon_s.value),
                &(procmon_s.value_old)) != 0)
    abort();
}
void procmon_dtor(
                  void
                  )
{
  /* ----------------------------------------------------- */
  /* reset old catcher after itimer in case timer goes off */
  /* ----------------------------------------------------- */
  if (setitimer(ITIMER_PROF, &(procmon_s.value_old),
                &(procmon_s.value)) != 0)
    abort();
  if (sigaction(SIGPROF, &(procmon_s.action_old),
                &(procmon_s.action)) != 0)
    abort();
  /* ------------------- */
  /* get sampling at end */
  /* ------------------- */
  procmon_handler(0,NULL,NULL);
  /* --------- */
  /* free info */
  /* --------- */
  if( procmon_info_free() != 0 )
    abort();
}
