/*
 * This code was originally obtained from
 * Brett Kettering(LANL) and Phil Mucci(UTK)
 */
#if !defined( PROCMON_H )
#define PROCMON_H
#if !defined( PROCMON_EXTERN )
#define PROCMON_EXTERN extern
#endif /* !defined( PROCMON_EXTERN ) */
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/procfs.h>
#include <sys/resource.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <assert.h>
#define PROCMON_MAXHOSTNAMELEN 256
/* ------------------------------------------------------ */
/* names of sample data                                   */
/* If you wish to have a copy of just these struct arrys: */
/*    #define PROCMON_EXTERN_OFF                          */
/* ------------------------------------------------------ */
typedef struct
{
  char *name;
  int value;
}
PROCMON_NAME_VALUE_struct;
/* ------------------------------------------------------------------------ */
/* equivalent data (apparently - tested on linux/sgi/osf servers)           */
/* ------------------------------------------------------------------------ */
/* user time                                                                */
/*   rusage_info.ru_utime == prusage_info.pu_utime == prusage_info.pr_utime */
/* system time                                                              */
/*   rusage_info.ru_stime == prusage_info.pu_stime == prusage_info.pr_stime */
/* wall clock time                                                          */
/*   glock_gettime == prusage_info.pu_tstamp+prusage_info.pu_starttime ==   */
/*                    prusage_info.pr_rtime                                 */
/* page reclaims, faults minor                                              */
/*   rusage_info.ru_minflt == prusage_info.pr_minf == prusage_info.pu_minf  */
/* page faults, faults major                                                */
/*   rusage_info.ru_majflt == prusage_info.pr_majf == prusage_info.pr_majf  */
/* swaps                                                                    */
/*   rusage_info.ru_nswap == prusage_info.pr_nswap == prusage_info.pu_nswap */
/* ------------------------------------------------------------------------ */
#define PROCMON_NAME_VALUE_MAX 132
#if !defined( PROCMON_EXTERN_OFF )
extern PROCMON_NAME_VALUE_struct procmon_name_value_int_vector[];
#else /* PROCMON_EXTERN_OFF */
PROCMON_NAME_VALUE_struct procmon_name_value_int_vector[PROCMON_NAME_VALUE_MAX] = {
  {"mallinfo         - Total Arena Space",                    0},
  {"mallinfo         - Number of Ordinary Blocks",            1},
  {"mallinfo         - Number of Small Blocks",               2},
  {"mallinfo         - Number of Holding Blocks",             3},
  {"mallinfo         - space in holding block headers",       4},
  {"mallinfo         - space in small blocks in use",         5},
  {"mallinfo         - space in free small blocks",           6},
  {"mallinfo         - space in ordinary blocks in use",      7},
  {"mallinfo         - space in free ordinary blocks",        8},
  {"mallinfo         - cost of enabling keep option",         9},
  {"getrusage        - maximum resident set size",           10},
  {"getrusage        - integral shared memory size",         11},
  {"getrusage        - integral unshared data size",         12},
  {"getrusage        - integral unshared stack size",        13},
  {"getrusage        - page reclaims",                       14},
  {"getrusage        - page faults",                         15},
  {"getrusage        - swaps",                               16},
  {"getrusage        - block input operations",              17},
  {"getrusage        - block output operations",             18},
  {"getrusage        - messages sent",                       19},
  {"getrusage        - messages received",                   20},
  {"getrusage        - signals received",                    21},
  {"getrusage        - voluntary context switches",          22},
  {"getrusage        - involuntary context switches",        23},
  {"ioctl:PIOCPSINFO - Pages Resident",                      24},
  {"ioctl:PIOCPSINFO - Pages Image",                         25},
  {"", 26}
};
#endif /* PROCMON_EXTERN_OFF */
#if !defined( PROCMON_EXTERN_OFF )
extern PROCMON_NAME_VALUE_struct procmon_name_value_float_vector[];
#else /* PROCMON_EXTERN_OFF */
PROCMON_NAME_VALUE_struct procmon_name_value_time_data[PROCMON_NAME_VALUE_MAX] = {
  {"clock_gettime    - Time_Absolute  ",                      0},
  {"clock_gettime    - Time_Wall      ",                      1},
  {"clock_gettime    - Time_Wall_Delta",                      2},
  {"getrusage        - Time_User      ",                      3},
  {"getrusage        - Time_System    ",                      4},
  {"ioctl:PIOCUSAGE  - Time_Wait      ",                      5},
  {"", 8}
};
#endif /* PROCMON_EXTERN_OFF */
/* -------------------------- */
/* DONE: names of sample data */
/* -------------------------- */
/* ----------------------------- */
/* structure holding sample data */
/* ----------------------------- */
typedef struct
{
  long long
    num_samples, /* number of samples */
    num_samples_interval, /* number of samples in this interval */
    num_samples_printed; /* number of samples printed */
  /* constants */
  long long
    page_size; /* size of a page */
  /* sample info [min,current,max] */
  unsigned long long
    int_vector[PROCMON_NAME_VALUE_MAX][3]; /* integral types */
  int
    int_vector_ndefined[PROCMON_NAME_VALUE_MAX]; /* if defined */
  /* scalar info */
  double
    time_data[PROCMON_NAME_VALUE_MAX]; /* time data */
  int
    time_data_ndefined[PROCMON_NAME_VALUE_MAX]; /* if defined */
}
PROCMON_SAMPLE_INFO_struct;
/* ---------------------------------------------- */
/* structure holding generic data and sample data */
/* ---------------------------------------------- */
typedef struct
{
  struct sigaction
    action, /* new action */
    action_ignore, /* when setting to ignore the signal */
    action_old; /* old action */
  struct itimerval
    value, /* new value */
    value_old; /* old value */
  char
    *filename, /* filename to print output */
    hostname[PROCMON_MAXHOSTNAMELEN]; /* gethostname() */
  FILE
    *filename_fp; /* file handle to output file */
  int
    filename_fd, /* file descriptor gotten from mkstemp command */
    initialized; /* if this has been initialized or not */
  pid_t
    pid; /* process id */
  long long
    millisecs, /* number of milliseconds between sampling */
    output_interval, /* output every output_interval samples */
    output_sample_min; /* minimum number of samples to keep file */
  PROCMON_SAMPLE_INFO_struct
    sample_info; /* collection of the sample data */
  time_t
    time_init_coarse; /* time() */
  struct timespec
    time_current, /* clock_gettime */
    time_init_fine; /* clock_gettime() */
}
PROCMON_struct;
PROCMON_EXTERN PROCMON_struct procmon_s;
/* ------------- */
/* useful macros */
/* ------------- */
#define PROCMON_MAX( x, y ) (((x) > (y)) ? (x) : (y))
#define PROCMON_MIN( x, y ) (((x) < (y)) ? (x) : (y))
#define PROCMON_TIMEN_TO_DOUBLE(t) \
  ((double)((t).tv_sec) + (double)((t).tv_nsec) / 1000000000.0)
#define PROCMON_TIMEU_TO_DOUBLE(t) \
  ((double)((t).tv_sec) + (double)((t).tv_usec) / 1000000.0)
/* ---------- */
/* prototypes */
/* ---------- */
#if defined( _PROCMON_CC_GCC )
/* ------------------------------------------------------------- */
/* define init/fini routines via attribute as opposed to ld flag */
/* ------------------------------------------------------------- */
void procmon_ctor( void ) __attribute__((constructor));
void procmon_dtor( void ) __attribute__((destructor));
/* ----------------------------------------------------------------------- */
/* I guess this was originally for user code to create their own routines  */
/* (named process_probe_*_routine) to overwrite (weak attribute) the ones  */
/* that exist in this lib (__(f|g) alias)???                               */
/* ----------------------------------------------------------------------- */
static void __f () { ; }
static void __g () { ; }
void process_probe_init_routine () __attribute__ ((weak, alias ("__f")));
void process_probe_shutdown_routine () __attribute__ ((weak, alias ("__g")));
#else
void procmon_ctor(
                  void
                  );
void procmon_dtor(
                  void
                  );
#endif
int procmon_info_init(
                      void
                      );
int procmon_info_free(
                      void
                      );
void procmon_handler(
                     int sig,
                     siginfo_t *info,
                     void *context
                     );
#endif /* !defined( PROCMON_H ) */
