#include <stdlib.h>
#include <malloc.h>
#include <signal.h>
#include <unistd.h>
#include <assert.h>
/* ------------- */
/* include files */
/* ------------- */
#if defined( _PROCMON_UPS )
#include "ups.h"
#else
#if defined( _PROCMON_serial )
#else
#include "mpi.h"
#endif
#endif

#define STRING_CHUNK_SIZE 4096
#define NUM_STRINGS 30
#define NUM_REPITITIONS 5000
int main(int argc, char **argv)
{
  int
    i, /* loop var */
    ierr = 0,
    j, /* loop var */
    k, /* loop var */
    penum=0; /* process number */
  char
    *p; /* leaky char array */

#if defined( _PROCMON_DEBUG_EXTRA )
  fprintf( stderr, "extra before init\n" );
#endif
  /* init and get process number */
#if defined( _PROCMON_UPS )
  ierr = UPS_AA_Init( argc, argv );
  assert( !ierr );
  ierr = UPS_CM_Get_penum( &penum );
  assert( !ierr );
#else
#if defined( _PROCMON_serial )
  penum = 0;
#else
  ierr = MPI_Init( &argc, &argv );
  assert( !ierr );
  ierr = MPI_Comm_rank( MPI_COMM_WORLD, &penum );
  assert( !ierr );
#endif
#endif

#if defined( _PROCMON_DEBUG_EXTRA )
  fprintf( stderr, "extra after init\n" );
#endif
  /* ------------------------------------ */
  /* send some user signals (at least 10) */
  /* only send signal if RLD set up       */
  /* ------------------------------------ */
  if( getenv( _PROCMON_RLD_NAME ) )
    {
      for( i=0; i<(penum%4)+10; i++ )
        {
          kill( getpid(), SIGPROF );
        }
    }
  /* --------------------------------------------------------- */
  /* loop doing allocation and some busy work to trigger timer */
  /* and create mem leak.                                      */
  /* --------------------------------------------------------- */
  for( i=0; i<NUM_STRINGS; i++ )
    {
      p = (char *)calloc( STRING_CHUNK_SIZE*(i+1), sizeof( char ) );
      for( k = 0; k < NUM_REPITITIONS; k++ )
        {
          for( j = 0; j < (i+1)*STRING_CHUNK_SIZE; j++ )
            {
              p[j] = '1';
            }
        }
      /* --------------------------------------- */
      /* some do not free - create leaky pointer */
      /* --------------------------------------- */
      if( penum != 0 )
        {
          free( p );
        }
    }
#if defined( _PROCMON_DEBUG_EXTRA )
  fprintf( stderr, "extra before terminate %d\n", penum );
#endif
  /* terminate */
#if defined( _PROCMON_UPS )
  ierr = UPS_AA_Terminate();
  assert( !ierr );
#else
#if defined( _PROCMON_serial )
  penum = 0;
#else
  ierr = MPI_Finalize();
  assert( !ierr );
#endif
#endif

#if defined( _PROCMON_DEBUG_EXTRA )
  fprintf( stderr, "extra after  terminate %d\n", penum );
#endif
  return( ierr );
}
